/*
 * Copyright  2005 Steve Whitley / NullFX Software
 * 
 * It is a violation of your purchase agreement to remove this text
 * This source may not be redistributed in any way shape or form
 * 
 */
using System;
using System.Collections.Generic;
using System.IO;
using System.Net;
using System.Net.Security;
using System.Net.Sockets;
using System.Reflection;
using System.Runtime.CompilerServices;
using System.Security.Cryptography.X509Certificates;
using System.Text;
using System.Text.RegularExpressions;
using System.Threading;
#if LOG4NET
using log4net;
#endif

namespace NullFX.Net {
    /// <summary>
    /// The FTP client used for all FTP client operations
    /// </summary>
    public sealed class FtpClient : IDisposable {
        private delegate SocketEx GetDataConnectionDelegate( object syncRoot );
#if LOG4NET
        private static readonly ILog log = LogManager.GetLogger( MethodBase.GetCurrentMethod( ).DeclaringType );
#endif
        //static readonly object SyncRoot = new object( );
        private SocketEx client;
        private readonly string hostValidationName;
        private readonly IPEndPoint endpoint;
        private readonly NetworkCredential credentials;
        private SecureState securityState;
        private RemoteCertificateValidationCallback serverValidation;
        private X509Certificate serverCert;
        private X509Chain serverChain;
        private SslPolicyErrors serverSslErrors;
        private readonly GetDataConnectionDelegate GetDataDataConnectionAsync;
        private ServerCertificateValidationMode serverCertMode = ServerCertificateValidationMode.AllowAny;
        private readonly string localFileDirectory;
        private string workingDirectory = "/";
        private string tempPortString;
        private bool abortOperation;
        private bool disposed;
        const int DEFAULT_BLOCK_LENGTH = 8192;

        /// <summary>
        /// An event fired when data is uploaded
        /// </summary>
        public event EventHandler<TransferProgressEventArgs> Uploaded;

        /// <summary>
        /// An event fired when data is downloaded
        /// </summary>
        public event EventHandler<TransferProgressEventArgs> Downloaded;

        /// <summary>
        /// Defines whether or not logging is used
        /// </summary>
        public bool EnableLogging { get; set; }

        /// <summary>
        /// Gets or sets whether the underlying connection uses an active or passive connection
        /// </summary>
        public ActivePassiveState ActivePassiveState { get; set; }

        /// <summary>
        /// If Connected to an FTP server, Gets the remote endpoint of the command connection
        /// </summary>
        public EndPoint RemoteCommandEndpoint {
            get {
                if ( client.Connected ) {
                    return client.RemoteEndPoint;
                }
                return endpoint;
            }
        }

        /// <summary>
        /// If Connected to an FTP server, Gets the local endpoint of the command connection (if not connected returns null)
        /// </summary>
        public EndPoint LocalCommandEndpoint {
            get {
                if ( client.Connected ) {
                    return client.LocalEndPoint;
                }
                return null;
            }
        }

        /// <summary>
        /// Gets or sets the current working directory
        /// </summary>
        public string WorkingDirectory {
            get {
                if ( client != null && client.Connected ) {
                    FtpResponse response = SendCommand( CommandType.PrintWorkingDirectory, "", 257 );
                    if ( response.IsValid ) {
                        Match m = Regex.Match( response.ResponseText, @"[""](?<path>\/.*)[""].*\r\n$", RegexOptions.ExplicitCapture );
                        if ( m.Groups["path"].Success ) {
                            workingDirectory = m.Groups["path"].Value;
                        }
                    }
                }
                return workingDirectory;
            }
            set {
                if ( string.IsNullOrEmpty( value ) ) throw new ArgumentException( "WorkingDirectory cannot be null or empty" );
                if ( value != workingDirectory ) {
                    if ( client.Connected ) {
                        ChangeWorkingDirectory( value );
                    } else {
                        workingDirectory = value;
                    }
                }
            }
        }

        /// <summary>
        /// Gets whether the ftp client is connected
        /// </summary>
        public bool Connected {
            get {
                if ( client == null ) return false;
                return client.Connected;
            }
        }

        /// <summary>
        /// Gets the clients credentials used to log into the ftp server
        /// </summary>
        public NetworkCredential ClientCredentials { get { return credentials; } }

        /// <summary>
        /// If a TLS connection has been established this will Get the command connection certificate
        /// </summary>
        public X509Certificate RemoteServerCertificate { get { return serverCert; } }

        /// <summary>
        /// If a TLS connection has been established this will Get the command connection certificate chain
        /// </summary>
        public X509Chain RemoteServerChain { get { return serverChain; } }

        /// <summary>
        /// If a TLS connection has been established this will Get any security policy errors incurred
        /// </summary>
        public SslPolicyErrors SecureAuthenticationErrors { get { return serverSslErrors; } }

        /// <summary>
        /// If the FTP Client uses a TLS connection, this sets the mode in which the client accepts and validates the servers certificate.
        /// </summary>
        public ServerCertificateValidationMode SecureServerCertificateValidationMode {
            get { return serverCertMode; }
            set { serverCertMode = value; }
        }

        /// <summary>
        /// If the ftp client uses a TLS connection, this flag is used to specify whether the client sends a certificate to the server, and if so, how the client goes about obtaining that certificate.
        /// </summary>
        /// <remarks>If UseCallbackToObtainCert is used, YOU MUST assign SecureLocalCertificateSelectionCallback else the TLS handshake will fail.</remarks>
        public ClientCertificateMode SecureClientCertificateMode {
            get { return client.ClientCertificateMode; }
            set { client.ClientCertificateMode = value; }
        }

        /// <summary>
        /// Contains the client certificate collection
        /// </summary>
        public X509CertificateCollection ClientCertificates {
            get { return client.ClientCertificates; }
        }

        /// <summary>
        /// this event allows you to validate a server's certificate if TLS is used
        /// </summary>
        public event RemoteCertificateValidationCallback ServerValidation {
            [MethodImpl( MethodImplOptions.Synchronized )]
            add { serverValidation = (RemoteCertificateValidationCallback)Delegate.Combine( serverValidation, value ); }
            [MethodImpl( MethodImplOptions.Synchronized )]
            remove { serverValidation = (RemoteCertificateValidationCallback)Delegate.Remove( serverValidation, value ); }
        }

        /// <summary>
        /// internal routing for the ServerValidation event.
        /// </summary>
        private void OnServerValidationRecieved( object sender, CertificateValidationEventArgs e ) {
            if ( sender == client.InternalSslStream ) {
                serverCert = e.Certificate;
                serverChain = e.Chain;
                serverSslErrors = e.Errors;
            }
            if ( serverValidation != null ) {
                e.Allow = serverValidation( sender, e.Certificate, e.Chain, e.Errors );
            } else {
                e.Allow = true;
            }
        }

        /// <summary>
        /// Method used to fire the Uploaded event
        /// </summary>
        /// <param name="e">transfer progress event information</param>
        private void OnUploaded( TransferProgressEventArgs e ) {
            if ( Uploaded != null && e != null && e.BytesTransferred > 0) Uploaded( this, e );
        }

        /// <summary>
        /// Method used to fire the Downloaded event
        /// </summary>
        /// <param name="e">transfer progress event information</param>
        private void OnDownloaded( TransferProgressEventArgs e ) {
            if ( Downloaded != null && e != null && e.BytesTransferred > 0 ) Downloaded( this, e );
        }

        /// <summary>
        /// Connects to the FTP server
        /// </summary>
        public void Connect( ) {
            Connect( NetworkSecurityFlags.StandardFtp );
        }

        /// <summary>
        /// Connects to the FTP server
        /// </summary>
        /// <param name="networkSecurity">type of connection to establish</param>
        /// <param name="remoteServerName">used when authenticating the remote server's certificate (this affects the results of SslPolicyErrors in the ServerValidation event)</param>
        public void Connect( NetworkSecurityFlags networkSecurity ) {
            if ( !Enum.IsDefined( typeof( NetworkSecurityFlags ), networkSecurity ) ) throw new ArgumentException( "networkSecurity value is not valid" );
            LogDebug( "Connecting to {0} in {1} mode", endpoint.ToString( ), networkSecurity );
            if ( networkSecurity == NetworkSecurityFlags.ImpliedTls ) {
                client.ConnectAndAuthenticate( endpoint, hostValidationName );
                securityState = SecureState.Secure;
            } else {
                client.Connect( endpoint );
            }
            FtpResponse response = GetServerResponse( 220 );
            if ( !response.IsValid ) {
                LogError( "FTP send an unexpected response {0}", response.ResponseText );
                var ex = new FtpOperationException( "Failed to read Server's Welcome Message" ) { AdditionalInfo = response };
                throw ex;
            }
            if ( networkSecurity == NetworkSecurityFlags.AuthTls ) {
                LogDebug( "Sending AUTH TLS" );
                response = SendCommand( CommandType.Auth, "TLS", 234 );
                LogDebug( "Server Response: {0}", response.ResponseText );
                if ( !response.IsValid ) throw new FtpAuthenticationException( "AUTH TLS Failed" );
                LogDebug( "Initiating TLS handshake" );
                client.AuthenticateAsClient( hostValidationName );
                securityState = SecureState.Secure;
                LogDebug( "TLS handshake successful" );
            }
            LogDebug( "sending user name" );
            response = SendCommand( CommandType.User, credentials.UserName, 331 );
            if ( !response.IsValid ) {
                LogError( response.ResponseText );
                throw new FtpAuthenticationException( response.ResponseText );
            }
            LogDebug( "sending password" );
            response = SendCommand( CommandType.Password, credentials.Password, 230 );
            if ( !response.IsValid ) {
                LogError( response.ResponseText );
                throw new FtpAuthenticationException( string.Format( "password invalid for user {0}", credentials.UserName ) );
            }
            if ( securityState == SecureState.Secure/* && networkSecurity == NetworkSecurityFlags.AuthTls*/) {
                LogDebug( "Setting protection level to private" );
                response = SendCommand( CommandType.Protection, ( (char)SecureProtectionLevel.Private ).ToString( ), 200 );
                if ( !response.IsValid ) {
                    LogError( response.ResponseText );
                    var ex = new FtpOperationException( "A secure connection has been established but cannot establish a secure data transfer protocol (server failed when sending PROT P)" ) { 
                        AdditionalInfo = response 
                    };
                    throw ex;
                }
            }
            if ( string.IsNullOrEmpty( workingDirectory ) && ( workingDirectory != "/" || workingDirectory != "." ) ) {
                LogDebug( "changing default directory to working directory" );
                ChangeWorkingDirectory( workingDirectory );
            }
        }

        /// <summary>
        /// Aborts all data connection activity
        /// </summary>
        public void Abort( ) {
            if ( SendCommand( CommandType.Abort, "", 226 ).IsValid ) {
                abortOperation = true;
            }
        }

        /// <summary>
        /// Gets whether or not a file of the given size can be placed on the server
        /// </summary>
        /// <param name="size">file size to allocate on the server</param>
        public bool CanStoreAFileThisSize( long size ) {
            return SendCommand( CommandType.Allocate, size.ToString( ), 200, 202 ).IsValid;
        }

        /// <summary>
        /// Changes the working directory
        /// </summary>
        /// <param name="path">new directory path</param>
        public bool ChangeWorkingDirectory( string path ) {
            if ( string.IsNullOrEmpty( path ) ) throw new ArgumentException( "path cannot be null or empty" );
            FtpResponse response = SendCommand( CommandType.ChangeWorkingDirectory, path, 250 );
            if ( response.IsValid ) {
                workingDirectory = path;
            }
            return response.IsValid;
        }

        /// <summary>
        /// Closes the ftp connection
        /// </summary>
        public void Close( ) {
            SendCommand( CommandType.Logout, "", 221 );
            if ( securityState == SecureState.Secure ) {
                client.Close( );
                securityState = SecureState.PlainText;
            }
            client.Close( );
            client = null;
            client = new SocketEx( );
        }

        /// <summary>
        /// Remove a file or directory from the server
        /// </summary>
        /// <param name="targetType">type of thing to remove</param>
        public bool Remove( RemoveTargetType targetType, string fileOrDirectory ) {
            if ( !Enum.IsDefined( typeof( RemoveTargetType ), targetType ) )
                throw new ArgumentException( "targetType is not a valid RemoveTargetType" );
            if ( string.IsNullOrEmpty( fileOrDirectory ) )
                throw new ArgumentException( "fileOrDirectory cannot be null or empty" );
            switch ( targetType ) {
                case RemoveTargetType.Directory: {
                    FtpResponse response = SendCommand( CommandType.RemoveDirectory, fileOrDirectory, 250 );
                    return response.IsValid;
                }
                case RemoveTargetType.File: {
                    FtpResponse response = SendCommand( CommandType.Delete, fileOrDirectory, 250 );
                    return response.IsValid;
                }
            }
            return false;
        }

        /// <summary>
        /// Downloads a file
        /// </summary>
        public byte[] Download( string remoteFile ) {
            if ( string.IsNullOrEmpty( remoteFile ) ) throw new ArgumentException( "remoteFile cannot be null or empty" );
            LogDebug( "downloading {0}", remoteFile );
            object syncRoot = new object( );
            SetTransferEncoding( TransferEncoding.Binary );
            Monitor.Enter( syncRoot );
            LogDebug( "Creating data connection" );
            IAsyncResult bcdcResult = BeginCreateDataConnection( syncRoot );
            Monitor.Wait( syncRoot );
            FtpResponse response = SendCommand( CommandType.Retrieve, remoteFile, 150 );
            if ( response.IsValid ) {
                SocketEx dataConnection = EndCreateDataConnection( bcdcResult );
                LogDebug( "Data connection established, reading file" );
                byte[] buffer = ReadDataConnection( dataConnection );
                response = GetServerResponse( 226 );
                if ( response.IsValid ) {
                    return buffer;
                } else {
                    LogError( "FTP Response: {0}", response.ResponseText );
                    throw new FtpOperationException( "FTP Response: " + response.ResponseText );
                }
            } else {
                LogError( "FTP Response: {0}", response.ResponseText );
                throw new FtpOperationException( "FTP Response: " + response.ResponseText );
            }
        }

        /// <summary>
        /// Downloads a file
        /// </summary>
        public long Download( string remoteFile, string localFilePath ) {
            if ( string.IsNullOrEmpty( remoteFile ) ) throw new ArgumentException( "remoteFile cannot be null or empty" );
            if ( string.IsNullOrEmpty( localFilePath ) ) throw new ArgumentException( "localFilePath cannot be null or empty" );
            if ( !Path.IsPathRooted( localFilePath ) ) localFilePath = Path.Combine( localFileDirectory, localFilePath );
            long len = -1;
            using ( var lf = new FileStream( localFilePath, FileMode.Create, FileAccess.Write, FileShare.None ) ) {
                byte[] buffer = Download( remoteFile );
                lf.Write( buffer, 0, buffer.Length );
                len = lf.Length;
            }
            return len;
        }

        /// <summary>
        /// Gets a list of files and directories on the
        /// </summary>
        public string[] GetFileSystemList( string mask, bool displayFullPaths ) {
            object syncRoot = new object( );
            LogDebug( "Getting file system list" );
            SetTransferEncoding( TransferEncoding.ASCII );
            Monitor.Enter( syncRoot );
            LogDebug( "Creating the data connection" );
            IAsyncResult result = BeginCreateDataConnection( syncRoot );
            Monitor.Wait( syncRoot );
            if ( string.IsNullOrEmpty( tempPortString ) ) {
                LogError( "Getting file system list" );
                throw new InvalidOperationException( "Failed to establish a remote data connection with the FTP server" );
            }
            LogDebug( "Getting file list" );
            FtpResponse response = SendCommand( CommandType.NameList, mask, 150 );
            SocketEx dataConnection = EndCreateDataConnection( result );
            LogDebug( "Reading server response" );
            byte[] listBytes = ReadDataConnection( dataConnection );
            response = GetServerResponse( 226 );
            if ( !response.IsValid ) {
                LogError( "FTP Server Responded: {0}", response.ResponseText );
                throw new FtpOperationException( "FTP Server Responded: " + response.ResponseText );
            }
            string[] list = Encoding.ASCII.GetString( listBytes ).Replace( "\r", "" ).Split( new string[] { "\n" }, StringSplitOptions.RemoveEmptyEntries );
            if ( !displayFullPaths ) {
                for ( int i = 0; i < list.Length; i++ ) {
                    list[i] = Path.GetFileName( list[i] );
                }
            }
            return list;
        }

        /// <summary>
        /// Gets a list of files and directories on the
        /// </summary>
        public string[] GetFileSystemList( string mask ) {
            return GetFileSystemList( mask, false );
        }

        /// <summary>
        /// Gets a list of files and directories on the
        /// </summary>
        public string[] GetFileSystemList( ) {
            return GetFileSystemList( "", false );
        }

        /// <summary>
        /// Creates a new directory on the server at the given path
        /// </summary>
        public bool CreateDirectory( string path ) {
            if ( string.IsNullOrEmpty( path ) ) throw new ArgumentException( "file cannot be null or empty" );
            return SendCommand( CommandType.MakeDirectory, path, 257 ).IsValid;
        }

        /// <summary>
        /// Gets the size of a remote file
        /// </summary>
        public long GetFileSize( string file ) {
            return GetFileSize( file, false );
        }

        /// <summary>
        /// Gets the size of a remote file
        /// </summary>
        public long GetFileSize( string file, bool throwOnError ) {
            if ( string.IsNullOrEmpty( file ) ) throw new ArgumentException( "file cannot  be null or empty" );
            FtpResponse response = SendCommand( CommandType.Size, file, 213 );
            if ( !response.IsValid ) {
                if ( throwOnError ) {
                    throw new FtpOperationException( "FTP Response: " + response.ResponseText );
                } else {
                    return -1;
                }
            }
            try {
                long rtn = long.Parse( response.ResponseText.Trim( ) );
                return rtn;
            } catch ( Exception e ) {
                if ( throwOnError ) {
                    throw new FtpOperationException( "Failure to parse ftp response of a SIZE operation", e );
                } else {
                    return -1;
                }
            }
        }

        /// <summary>
        /// Upload a file to the server
        /// </summary>
        public long Upload( Stream file, string remoteFileName ) {
            if ( file == null ) throw new ArgumentException( "File cannot be null" );
            if ( string.IsNullOrEmpty( remoteFileName ) ) throw new ArgumentException( "remote file cannot be null or empty" );
            byte[] buffer = new byte[file.Length];
            file.Read( buffer, 0, buffer.Length );
            return Upload( buffer, remoteFileName );
        }

        /// <summary>
        /// Upload a file to the server
        /// </summary>
        public long Upload( byte[] file, string remoteFileName ) {
            if ( file == null ) throw new ArgumentException( "File cannot be null" );
            if ( string.IsNullOrEmpty( remoteFileName ) ) throw new ArgumentException( "remote file cannot be null or empty" );
            return Upload( file, 0, file.Length, remoteFileName );
        }

        /// <summary>
        /// Upload a file to the server
        /// </summary>
        public long Upload( string localFile, string remoteFileName ) {
            if ( string.IsNullOrEmpty( localFile ) ) throw new ArgumentException( "localFile cannot be null or empty" );
            if ( string.IsNullOrEmpty( remoteFileName ) ) throw new ArgumentException( "remote file cannot be null or empty" );
            if ( !Path.IsPathRooted( localFile ) ) localFile = Path.Combine( localFileDirectory, localFile );
            if ( !File.Exists( localFile ) ) throw new ArgumentException( localFile + " file does not exist" );
            using ( Stream s = File.OpenRead( localFile ) ) {
                return Upload( s, remoteFileName );
            }
        }

        /// <summary>
        /// Upload a file to the server
        /// </summary>
        public long Upload( byte[] file, int start, int count, string remoteFileName ) {
            if ( file == null ) throw new ArgumentException( "File cannot be null" );
            if ( string.IsNullOrEmpty( remoteFileName ) ) throw new ArgumentException( "remote file cannot be null or empty" );
            LogDebug( "uploading {0} file", remoteFileName );
            object syncRoot = new object( );
            SetTransferEncoding( TransferEncoding.Binary );
            Monitor.Enter( syncRoot );
            IAsyncResult bcdResult = BeginCreateDataConnection( syncRoot );
            Monitor.Wait( syncRoot );
            FtpResponse response = SendCommand( CommandType.Store, remoteFileName, 150 );
            if ( response.IsValid ) {
                SocketEx dataConnection = EndCreateDataConnection( bcdResult );
                LogDebug( "Writing data to server" );
                long len = WriteDataConnection( dataConnection, file, start, count );
                response = GetServerResponse( 226 );
                if ( response.IsValid ) {
                    LogDebug( "{0} uploaded successfully", remoteFileName );
                    return len;
                } else {
                    LogError( "FTP Response: ", response.ResponseText );
                    throw new FtpOperationException( "FTP Response: " + response.ResponseText );
                }
            } else {
                LogError( "FTP Response: ", response.ResponseText );
                throw new FtpOperationException( "FTP Response: " + response.ResponseText );
            }
        }

        /// <summary>
        /// Renames a file or directory
        /// </summary>
        public bool Rename( string from, string to ) {
            if ( string.IsNullOrEmpty( from ) ) throw new ArgumentException( "from cannot be null or empty" );
            if ( string.IsNullOrEmpty( to ) ) throw new ArgumentException( "to cannot be null or empty" );
            FtpResponse response = SendCommand( CommandType.RenameFrom, from, 350 );
            if ( response.IsValid ) {
                response = SendCommand( CommandType.RenameTo, to, 250 );
                return response.IsValid;
            }
            return false;
        }

#if SERVER_FEAT_IMPLEMENTED
        /// <summary>
        /// Gets a list of features the ftp server advertises
        /// </summary>
        private static void GetServerFeatureSet( ) {
            throw new System.NotImplementedException( );
        }
#endif
        /// <summary>
        /// send an ftp command to the server
        /// </summary>
        private FtpResponse SendCommand( CommandType commandType, string args, params short[] expectedReturnCodes ) {
            return SendCommand( new FtpCommand( commandType, args, expectedReturnCodes ) );
        }

        /// <summary>
        /// send an ftp command to the server
        /// </summary>
        private FtpResponse SendCommand( FtpCommand command ) {
            if ( !Connected ) throw new FtpOperationException( "your connection has either been lost or has not been established" );
            if ( EnableLogging ) {
                string commandString = command.ToString( );
                if ( command.CommandType == CommandType.User ) {
                    commandString = "USER *******";
                }
                if ( command.CommandType == CommandType.Password ) {
                    commandString = "PASS *******";
                }
                commandString = commandString.TrimEnd( '\r', '\n' );
                LogDebug( "Client Sending:  {0}", commandString );
            }
            client.Send( command );
            return GetServerResponse( command.ExpectedReturnCodes );
        }

        /// <summary>
        /// reads the server response and returns an ftp response
        /// </summary>
        private FtpResponse GetServerResponse( params short[] expectedResponseCodes ) {
            if ( !Connected ) throw new FtpOperationException( "your connection has either been lost or has not been established" );
            List<byte> finalBytes = new List<byte>( );
            byte[] buffer = new byte[8192];
            byte[] temp;
            int count = 0;
            bool doContinue = true;
            do {
                count = client.Receive( buffer );
                temp = new byte[count];
                Buffer.BlockCopy( buffer, 0, temp, 0, count );
                finalBytes.AddRange( temp );
                string tempStr = Encoding.ASCII.GetString( temp );
                if ( count != buffer.Length && FtpResponse.CompleteLine.IsMatch( tempStr ) ) {
                    doContinue = false;
                }
            } while ( doContinue );
            string rawResponse = Encoding.ASCII.GetString( finalBytes.ToArray( ) );
            var loggingResponse = rawResponse.TrimEnd( 'r', '\n' );
            LogDebug( "Server Response: {0}", loggingResponse );
            return new FtpResponse( rawResponse, expectedResponseCodes );
        }

        /// <summary>
        /// Asynchronous Create data connection method
        /// </summary>
        /// <returns>the IAsyncResult of the operation</returns>
        private IAsyncResult BeginCreateDataConnection( object syncRoot ) {
            return GetDataDataConnectionAsync.BeginInvoke( syncRoot, null, null );
        }

        /// <summary>
        /// Asynchronous Create data connection method
        /// </summary>
        /// <param name="result">the IAsyncResult from BeginCreateDataConnection</param>
        /// <returns>the socket created</returns>
        internal SocketEx EndCreateDataConnection( IAsyncResult result ) {
            return GetDataDataConnectionAsync.EndInvoke( result );
        }

        internal SocketEx CreateDataConnection( ) {
            object locker = new object( );
            return CreateDataConnection( locker );
        }

        /// <summary>
        /// creates a data socket connection
        /// </summary>
        private SocketEx CreateDataConnection( object syncRoot ) {
            tempPortString = null;
            Monitor.Enter( syncRoot );
            SocketEx socket = new SocketEx( );
            IPEndPoint endpoint = null;
            FtpResponse response;
            LogDebug( "Creating data connection in {0} mode", ActivePassiveState );
            if ( ActivePassiveState == ActivePassiveState.Active ) {
                IPAddress address = IPAddress.Any;
                LogDebug( "Resolving IP Address for {0}", Environment.MachineName );
                foreach ( IPAddress test in Dns.GetHostEntry( Environment.MachineName ).AddressList ) {
                    if ( test.AddressFamily == socket.AddressFamily ) {
                        address = test;
                        LogDebug( "{0} resolved to {1}", Environment.MachineName, test.ToString( ) );
                        break;
                    }
                }
                endpoint = new IPEndPoint( address, 0 );
                socket.Bind( endpoint );
                socket.Listen( int.MaxValue );
                IAsyncResult acceptEndResult = socket.BeginAccept( null, null );
                tempPortString = string.Format( "{0},{1},{2}", ( (IPEndPoint)socket.LocalEndPoint ).Address.ToString( ).Replace( ".", "," ), ( (IPEndPoint)socket.LocalEndPoint ).Port >> 8, ( (IPEndPoint)socket.LocalEndPoint ).Port & 0x00ff );
                LogDebug( "listening at {0} ({1})", endpoint.ToString( ), tempPortString );
                response = SendCommand( CommandType.Port, tempPortString, 200 );
                if ( response.IsValid ) {
                    Monitor.Pulse( syncRoot );
                    Monitor.Exit( syncRoot );
                    SocketEx dataConnection = socket.EndAccept( acceptEndResult );
                    LogDebug( "Data connection with FTP server established" );
                    if ( securityState == SecureState.Secure ) {
                        try {
                            LogDebug( "Starting TLS handshake for data connection" );
                            dataConnection.AuthenticateAsClient( hostValidationName );
                            LogDebug( "TLS handshake successful" );
                        } catch ( Exception e ) {
                            LogError( e );
                        }
                    }
                    socket.Close( );
                    socket = null;
                    return dataConnection;
                } else {
                    LogError( "ERROR: {0}", response.RawResponse );
                }
            } else {
                response = SendCommand( CommandType.Passive, "", 227 );
                if ( response.IsValid ) {
                    Match m = Regex.Match( response.ResponseText, @"[(](?<ip>\d{1,3}[,]\d{1,3}[,]\d{1,3}[,]\d{1,3})[,](?<high>\d+)[,](?<low>\d+)[)]", RegexOptions.ExplicitCapture );
                    if ( m.Success ) {
                        IPAddress addr = IPAddress.Parse( m.Groups["ip"].Value.Replace( ",", "." ) );
                        int port = ( int.Parse( m.Groups["high"].Value ) << 8 ) | int.Parse( m.Groups["low"].Value );
                        endpoint = new IPEndPoint( addr, port );
                        tempPortString = endpoint.ToString( );
                        LogDebug( "connecting to {0} ({1},{2},{3}) for data connection", endpoint.ToString( ), m.Groups["ip"].Value, m.Groups["high"].Value, m.Groups["low"].Value );
                        IAsyncResult connectResult = null;
                        if ( securityState == SecureState.Secure ) {
                            connectResult = socket.BeginConnectAndAuthenticate( endpoint, hostValidationName, null, null );
                        } else {
                            connectResult = socket.BeginConnect( endpoint, null, null );
                        }
                        Monitor.Pulse( syncRoot );
                        Monitor.Exit( syncRoot );
                        try {
                            if ( securityState == SecureState.Secure ) {
                                socket.EndConnectAndAuthenticate( connectResult );
                                LogDebug( "finished connecting and TLS authentication" );
                            } else {
                                socket.EndConnect( connectResult );
                                LogDebug( "finished connecting" );
                            }
                        } catch ( Exception e ) {
                            LogError( e );
                        }
                        return socket;
                    }
                } else {
                    LogError( "Couldn't establish passive connection: {0}", response.RawResponse );
                    Monitor.Pulse( syncRoot );
                    Monitor.Exit( syncRoot );
                }
            }
            throw new FtpOperationException( "Failed to establish a data connection" );
        }

        /// <summary>
        /// reads a data connection and returns the resulting bytes
        /// </summary>
        internal byte[] ReadDataConnection( SocketEx dataSocket ) {
            LogDebug( "Reading from data connection" );
            List<byte> finalBytes = new List<byte>( );
            byte[] buffer = new byte[DEFAULT_BLOCK_LENGTH];
            byte[] temp;
            int count = 0;
            bool doContinue = true;
            do {
                try {
                    count = dataSocket.Receive( buffer );
                    OnDownloaded( new TransferProgressEventArgs( count ) );
                } catch ( IOException ioe ) {
                    LogError( ioe );
                }
                if ( count != buffer.Length ) {
                    temp = new byte[count];
                    Buffer.BlockCopy( buffer, 0, temp, 0, count );
                    buffer = temp;
                    doContinue = count > 0;
                }
                finalBytes.AddRange( buffer );
            } while ( doContinue && !abortOperation );
            dataSocket.Close( );
            if ( abortOperation ) abortOperation = false;
            LogDebug( "Read {0} bytes", finalBytes.Count );
            return finalBytes.ToArray( );
        }

        /// <summary>
        /// writes a set of bytes to a data connection
        /// </summary>
        private long WriteDataConnection( SocketEx dataSocket, byte[] buffer, int start, int count ) {
            LogDebug( "Writing to data connection" );
            byte[] bytes = start != 0 && count != buffer.Length ? new byte[count - start] : buffer;
            long sent = bytes.Length;
            if ( bytes[0] != buffer[0] ) {
                Buffer.BlockCopy( buffer, start, bytes, 0, bytes.Length );
            }
            if ( securityState == SecureState.Secure ) {
                dataSocket.AuthenticateAsClient( hostValidationName );
            }
            if ( bytes.Length < DEFAULT_BLOCK_LENGTH ) {
                dataSocket.Send( bytes, 0, bytes.Length, SocketFlags.None );
                OnUploaded( new TransferProgressEventArgs( bytes.Length ) );
            } else {
                byte[] temp = new byte[DEFAULT_BLOCK_LENGTH];
                int pos = 0;
                int len = temp.Length;
                do {
                    if ( pos + len > temp.Length ) len = temp.Length - pos;
                    Buffer.BlockCopy( bytes, pos, temp, 0, len );
                    pos += temp.Length;
                    sent = pos;
                    dataSocket.Send( temp, 0, len, SocketFlags.None );
                    OnUploaded( new TransferProgressEventArgs( len ) );
                } while ( pos < bytes.Length && !abortOperation );
            }
            dataSocket.Close( );
            if ( abortOperation ) abortOperation = false;
            LogDebug( "{0} bytes written", sent );
            return sent;
        }

        /// <summary>
        /// sets whether ASCII or Image (binary) mode is used
        /// </summary>
        public void SetTransferEncoding( TransferEncoding transferEncoding ) {
            LogDebug( "Setting transfer encoding to {0}", transferEncoding );
            if ( !SendCommand( CommandType.Type, ( (char)transferEncoding ).ToString( ), 200 ).IsValid ) {
                LogError( "Could not switch the encoding type to {0}", transferEncoding );
                throw new FtpOperationException( string.Format( "Could not switch the encoding type to {0}", transferEncoding ) );
            }
        }

        /// <summary>
        /// disposes the FTP client
        /// </summary>
        public void Dispose( ) {
            if ( !disposed ) {
                client.Close( );
                client.Dispose( );
                GC.SuppressFinalize( this );
                disposed = true;
            }
        }

        /// <summary>
        /// disposes the FTP Client
        /// </summary>
        void IDisposable.Dispose( ) {
            Dispose( );
        }

        private void LogDebug( string format, params object[] parameters ) {
#if LOG4NET
            if ( EnableLogging ) log.DebugFormat( format, parameters );
#else
            if ( EnableLogging ) Console.WriteLine( format, parameters );
#endif
        }

#if USE_WARN_INFO_LOGGING
        private void LogWarn( string format, params object[] parameters ) {
#if LOG4NET
            if ( EnableLogging ) log.WarnFormat( format, parameters );
#else
            if ( EnableLogging ) Console.WriteLine( "Warning: {0}", string.Format( format, parameters ) );
#endif
        }
#endif

        private void LogError( string format, params object[] parameters ) {
#if LOG4NET
            if ( EnableLogging ) log.ErrorFormat( format, parameters );
#else
            if ( EnableLogging ) Console.WriteLine( "Error: {0}", string.Format( format, parameters ) );
#endif
        }

        private void LogError( Exception ex ) {
#if LOG4NET
            if ( EnableLogging ) log.Error(message, ex);
#else
            if ( EnableLogging ) Console.WriteLine( "Error: {0}", ex );
#endif
        }

#if USE_WARN_INFO_LOGGING
        private void LogInfo( string format, params object[] parameters ) {
#if LOG4NET
            if ( EnableLogging ) log.InfoFormat( format, parameters );
#else
            if ( EnableLogging ) Console.WriteLine( "Info: {0}", string.Format( format, parameters ) );
#endif
        }
#endif

        /// <summary>
        /// Creates a new FTP Client
        /// </summary>
        public FtpClient( )
            : this( new IPEndPoint( IPAddress.Loopback, 21 ), new NetworkCredential( "anonymous", "anonymous@no-domain.com" ), "local files" ) { }

        /// <summary>
        /// Creates a new FTP Client
        /// </summary>
        public FtpClient( IPEndPoint endpoint, NetworkCredential credentials )
            : this( endpoint, credentials, Path.Combine( Path.GetDirectoryName( Assembly.GetExecutingAssembly( ).Location ), "local files" ) ) { }

        /// <summary>
        /// Creates a new FTP Client
        /// </summary>
        public FtpClient( IPEndPoint endpoint, NetworkCredential credentials, string localFileDirectory ) {
            this.endpoint = endpoint;
            this.credentials = credentials;
            this.localFileDirectory = localFileDirectory;
            if ( !Directory.Exists( localFileDirectory ) ) {
                Directory.CreateDirectory( localFileDirectory );
            }
            hostValidationName = Dns.GetHostEntry( endpoint.Address ).HostName;
            ActivePassiveState = ActivePassiveState.Passive;
            securityState = SecureState.PlainText;
            GetDataDataConnectionAsync = CreateDataConnection;
            client = new SocketEx( );
            client.ValidateServerCertificate += OnServerValidationRecieved;
        }
    }
}
