/*
 * Copyright  2005 Steve Whitley / NullFX Software
 * 
 * It is a violation of your purchase agreement to remove this text
 * This source may not be redistributed in any way shape or form
 * 
 */
using System.Text;
using System.Collections.Generic;

namespace NullFX.Net {
    /// <summary>
    /// Represents an abstracted FTP command
    /// </summary>
    public struct FtpCommand {
        public const string CRLF = "\r\n";
        private CommandType commandType;
        private string args;
        private short[] expectedCodes;
        private static readonly Dictionary<CommandType, string> CommandTranslation = new Dictionary<CommandType, string> {
               { CommandType.Abort, "ABOR" },
               { CommandType.Account, "ACCT" },
               { CommandType.Allocate, "ALLO" },
               { CommandType.Append, "APPE" },
               { CommandType.Auth, "AUTH" },
               { CommandType.ChangeWorkingDirectory, "CWD" },
               { CommandType.Delete, "DELE" },
               { CommandType.Features, "FEAT" },
               { CommandType.Help, "HELP" },
               { CommandType.List, "LIST" },
               { CommandType.Logout, "QUIT" },
               { CommandType.MakeDirectory, "MKD" },
               { CommandType.Mode, "MODE" },
               { CommandType.NameList, "NLST" },
               { CommandType.Noop, "NOOP" },
               { CommandType.Passive, "PASV" },
               { CommandType.Password, "PASS" },
               { CommandType.Port, "PORT" },
               { CommandType.PrintWorkingDirectory, "PWD" },
               { CommandType.Protection, "PROT" },
               { CommandType.Reinitialize, "REIN" },
               { CommandType.RemoveDirectory, "RMD" },
               { CommandType.RenameFrom, "RNFR" },
               { CommandType.RenameTo, "RNTO" },
               { CommandType.Restart, "REST" },
               { CommandType.Retrieve, "RETR" },
               { CommandType.SiteParameters, "SITE" },
               { CommandType.Size, "SIZE" },
               { CommandType.Status, "STAT" },
               { CommandType.Store, "STOR" },
               { CommandType.StoreUnique, "STOU" },
               { CommandType.StructureMount, "SMNT" },
               { CommandType.System, "SYST" },
               { CommandType.Type, "TYPE" },
               { CommandType.User, "USER" }
        };

        /// <summary>
        /// used to cache the command bytes
        /// </summary>
        private byte[] cache;

        /// <summary>
        /// FTP command
        /// </summary>
        public CommandType CommandType { get { return commandType; } set { commandType = value; } }

        /// <summary>
        /// The command parameters
        /// </summary>
        public string CommandArgs { get { return args; } set { args = value; } }

        /// <summary>
        /// valid FTP response codes
        /// </summary>
        public short[] ExpectedReturnCodes { get { return expectedCodes; } set { expectedCodes = value; } }

        /// <summary>
        /// Creates a new FTP Command object
        /// </summary>
        /// <param name="commandType">the ftp command type</param>
        /// <param name="args">any args that the command takes</param>
        public FtpCommand( CommandType commandType, string args ) : this( commandType, args, null ) { }

        /// <summary>
        /// Creates a new FTP Command object
        /// </summary>
        /// <param name="commandType">the ftp command type</param>
        /// <param name="args">any args that the command takes</param>
        /// <param name="expectedReturnCodes">any response codes that are to be considered to be valid</param>
        public FtpCommand( CommandType commandType, string args, short[] expectedReturnCodes ) {
            this.commandType = commandType;
            this.args = args;
            expectedCodes = expectedReturnCodes;
            cache = null;
        }

        /// <summary>
        /// returns the string format of the FTP command (one that the FTP Server will understand)
        /// </summary>
        public override string ToString( ) {
            StringBuilder sb = new StringBuilder( );
            sb.Append( CommandTranslation[commandType] );
            if ( !string.IsNullOrEmpty( args ) ) {
                sb.Append( " " );
                sb.Append( args );
            }
            sb.Append( CRLF );
            return sb.ToString( );
        }

        /// <summary>
        /// Converts the abstracted command into the byte buffer used to send to the FTP server
        /// </summary>
        /// <param name="command">the abstracted FTP representation of the command</param>
        /// <returns>the byte array buffer used to write to the socket</returns>
        public static implicit operator byte[]( FtpCommand command ) {
            if ( command.cache == null ) {
                command.cache = Encoding.ASCII.GetBytes( command.ToString( ) );
            }
            return command.cache;
        }
    }
}
