/*
 * Copyright  2005 Steve Whitley / NullFX Software
 * 
 * It is a violation of your purchase agreement to remove this text
 * This source may not be redistributed in any way shape or form
 * 
 */
using System;
using System.Text;
using System.Text.RegularExpressions;

namespace NullFX.Net {
    /// <summary>
    /// Represents an abstracted FTP server response
    /// </summary>
    public struct FtpResponse {
        /// <summary>
        /// An enumeration representing the validation result
        /// </summary>
        public enum ResponseValidation : byte { Unknown, Invalid, Valid }

        private ResponseValidation validation;
        private short responseCode;
        private short[] expectedCodes;
        private string rawResponse;
        private string responseText;
        private readonly Regex linePattern;
        private readonly static Regex completeLineRegex = new Regex( @"\d{3}\s.+\r\n$", RegexOptions.Singleline | RegexOptions.ExplicitCapture );

        /// <summary>
        /// Gets the Complete Line regular expression used for validation
        /// </summary>
        public static Regex CompleteLine { get { return completeLineRegex; } }

        /// <summary>
        /// Gets whether or not the response received was a valid one
        /// </summary>
        public bool IsValid {
            get {
                if(expectedCodes == null) return false;
                if(responseCode == (short)ResponseValidation.Unknown || responseCode == (short)ResponseValidation.Invalid) {
                    return false;
                }
                if(validation == ResponseValidation.Valid) return true;
                if(TestCode(responseCode, expectedCodes, ref validation)) {
                    return true;
                }
                return false;
            }
        }

        /// <summary>
        /// the response code (ex: 200)
        /// </summary>
        public short ResponseCode { get { return responseCode; } set { responseCode = value; } }

        /// <summary>
        /// the raw FTP server response (ex: "200 OK")
        /// </summary>
        public string RawResponse { get { return rawResponse; } set { rawResponse = value; } }

        /// <summary>
        /// the response text (ex: "OK")
        /// </summary>
        public string ResponseText { get { return responseText; } set { responseText = value; } }

        /// <summary>
        /// tests the received code against a few heuristics to determine whether or not the response was valid
        /// </summary>
        private static bool TestCode(short code, short[] expectedCodes, ref ResponseValidation validation) {
            if(expectedCodes == null || !(expectedCodes.Length > 0)) return false;
            if(validation == ResponseValidation.Valid) return true;
            if(validation == ResponseValidation.Invalid) return false;
            if(validation == ResponseValidation.Unknown && expectedCodes != null && expectedCodes.Length > 0) {
                foreach(short s in expectedCodes) {
                    if(s == code) {
                        validation = ResponseValidation.Valid;
                        return true;
                    }
                }
            }
            return false;
        }

        /// <summary>
        /// Creates a new FTP Response
        /// </summary>
        /// <param name="rawResponse">the raw FTP response string (ex: "200 OK")</param>
        /// <param name="expectedResponseCodes">list of expected valid response codes used for validation</param>
        public FtpResponse(string rawResponse, params short[] expectedResponseCodes) {
            this.rawResponse = rawResponse;
            validation = ResponseValidation.Unknown;
            expectedCodes = expectedResponseCodes;
            linePattern = new Regex(@"^(?<code>\d{3})(\s*(?<multiline>\-)?(?<message>.*))?$", RegexOptions.Multiline | RegexOptions.ExplicitCapture);
            short code = 0;
            StringBuilder message = new StringBuilder();
            foreach(string line in rawResponse.Split(new string[] { FtpCommand.CRLF }, StringSplitOptions.RemoveEmptyEntries)) {
                Match m = linePattern.Match(line);
                if(m.Success) {
                    if(code == 0 && m.Groups["code"].Success) {
                        code = short.Parse(m.Groups["code"].Value);
                        if(TestCode(code, expectedResponseCodes, ref validation)) {
                            responseCode = code;
                        }
                    }
                    if(m.Groups["message"].Success) {
                        message.Append(m.Groups["message"]);
                    }
                    if(m.Groups["multiline"].Success) {
                        message.Append(" ");
                    }
                }
            }
            responseCode = code;
            responseText = message.ToString();
        }

        /// <summary>
        /// Creates a new FTP Response
        /// </summary>
        /// <param name="responseCode">Response code returned from the server</param>
        /// <param name="responseText">text returned from the server</param>
        /// <param name="expectedResponseCodes">list of expected valid response codes used for validation</param>
        public FtpResponse(short responseCode, string responseText, params short[] expectedResponseCodes)
            : this(string.Format("{0} {1}\r\n", responseCode, responseText), expectedResponseCodes) { }

        /// <summary>
        /// returns the raw FTP server response (ex: "200 OK")
        /// </summary>
        public override string ToString() {
            return rawResponse;
        }

        /// <summary>
        /// The operation which converts the byte buffer into an abstracted FTP response object
        /// </summary>
        /// <param name="buffer">the result buffer</param>
        /// <returns>the abstracted FTP response object</returns>
        public static implicit operator FtpResponse(byte[] buffer) {
            return new FtpResponse(Encoding.ASCII.GetString(buffer));
        }
    }
}
