/*
 * Copyright  2005 Steve Whitley / NullFX Software
 * 
 * It is a violation of your purchase agreement to remove this text
 * This source may not be redistributed in any way shape or form
 * 
 */
using System;
using System.Net;
using System.Net.Security;
using System.Net.Sockets;
using System.Security.Authentication;
using System.Security.Cryptography.X509Certificates;

namespace NullFX.Net {
    /// <summary>
    /// An extended socket capable of plain text or SSL/TLS level security
    /// </summary>
    internal class SocketEx : IDisposable {
        private delegate SocketEx AcceptDelegate( );
        private delegate void ConnectDelegate( EndPoint endpoint );
        private delegate void ConnectAndAuthenticateDelegate( EndPoint endpoint, string targetHost );
        private Socket sock;
        private SslStream sslStream;
        private readonly AcceptDelegate AcceptInvoker;
        private readonly ConnectDelegate ConnectInvoker;
        private readonly ConnectAndAuthenticateDelegate ConnectAndAuthenticateInvoker;
        private readonly X509CertificateCollection clientCertificates = new X509CertificateCollection( );
        private ServerCertificateValidationMode validationMode = ServerCertificateValidationMode.AllowAny;
        private ClientCertificateMode clientCertMode = ClientCertificateMode.None;
        private bool secure;
        private bool disposed;

        public event EventHandler<CertificateValidationEventArgs> ValidateServerCertificate;

        public AddressFamily AddressFamily { get { return sock.AddressFamily; } }

        public bool Connected { get { return sock.Connected; } }

        public EndPoint LocalEndPoint { get { return sock.LocalEndPoint; } }

        public EndPoint RemoteEndPoint { get { return sock.RemoteEndPoint; } }

        public X509CertificateCollection ClientCertificates { get { return clientCertificates; } }

        public ClientCertificateMode ClientCertificateMode { get { return clientCertMode; } set { clientCertMode = value; } }

        public ServerCertificateValidationMode ServerCertificateValidationMode { get { return validationMode; } set { validationMode = value; } }

        public LocalCertificateSelectionCallback LocalCertificateSelectionCallback { get; set; }

        public X509Certificate RemoteCertificate {
            get {
                if ( sslStream == null ) return null;
                return sslStream.RemoteCertificate;
            }
        }

        public X509Certificate LocalCertificate {
            get {
                if ( sslStream == null ) return null;
                return sslStream.LocalCertificate;
            }
        }

        public bool IsAuthenticated {
            get {
                if ( sslStream == null ) return false;
                return sslStream.IsAuthenticated;
            }
        }

        public bool IsEncrypted {
            get {
                if ( sslStream == null ) return false;
                return sslStream.IsEncrypted;
            }
        }

        public bool IsSigned {
            get {
                if ( sslStream == null ) return false;
                return sslStream.IsSigned;
            }
        }

        public bool IsMutuallyAuthenticated {
            get {
                if ( sslStream == null ) return false;
                return sslStream.IsMutuallyAuthenticated;
            }
        }

        public SslStream InternalSslStream { get { return sslStream; } }

        public Socket InternalSocket { get { return sock; } }

        protected virtual void OnValidateServerCertificate( object sender, CertificateValidationEventArgs e ) {
            if ( ValidateServerCertificate != null ) ValidateServerCertificate( sender, e );
        }

        protected virtual X509Certificate DoLocalCertificateSelection( object sender, string targetHost, X509CertificateCollection localCertificates, X509Certificate remoteCertificate, string[] acceptableIssuers ) {
            if ( LocalCertificateSelectionCallback != null ) return LocalCertificateSelectionCallback( sender, targetHost, localCertificates, remoteCertificate, acceptableIssuers );
            return null;
        }

        private bool DoServerCertificateValidation( object sender, X509Certificate certificate, X509Chain chain, SslPolicyErrors sslPolicyErrors ) {
            switch ( validationMode ) {
                case ServerCertificateValidationMode.AllowAny: {
                    if ( ( sslPolicyErrors & SslPolicyErrors.RemoteCertificateNotAvailable ) == SslPolicyErrors.RemoteCertificateNotAvailable ) {
                        return false;
                    } else {
                        return true;
                    }
                }
                case ServerCertificateValidationMode.ManualValidationThroughValidationEvent: {
                    CertificateValidationEventArgs args = new CertificateValidationEventArgs( certificate, chain, sslPolicyErrors );
                    OnValidateServerCertificate( sender, args );
                    return args.Allow;
                }
                case ServerCertificateValidationMode.RequireMatchingNames: {
                    if ( ( sslPolicyErrors & SslPolicyErrors.RemoteCertificateNameMismatch ) == SslPolicyErrors.RemoteCertificateNameMismatch ) {
                        return false;
                    } else {
                        return true;
                    }
                }
                case ServerCertificateValidationMode.RequireNoChainErrors: {
                    if ( ( sslPolicyErrors & SslPolicyErrors.RemoteCertificateChainErrors ) == SslPolicyErrors.RemoteCertificateChainErrors ) {
                        return false;
                    } else {
                        return true;
                    }
                }
                case ServerCertificateValidationMode.RequireMatchingNames | ServerCertificateValidationMode.RequireNoChainErrors: {
                    if ( ( ( sslPolicyErrors & SslPolicyErrors.RemoteCertificateChainErrors ) == SslPolicyErrors.RemoteCertificateChainErrors ) ||
                       ( ( sslPolicyErrors & SslPolicyErrors.RemoteCertificateNameMismatch ) == SslPolicyErrors.RemoteCertificateNameMismatch ) ) {
                        return false;
                    } else {
                        return true;
                    }
                }
            }
            return false;
        }

        public void AuthenticateAsClient( string targetHost ) {
            AuthenticateAsClient( targetHost, new X509CertificateCollection( ), SslProtocols.Default, false );
        }

        public void AuthenticateAsClient( string targetHost, X509CertificateCollection clientCertificates, SslProtocols enabledProtocols, bool checkCertRevocation ) {
            CheckSslStream( );
            if ( !sock.Connected ) return;
            if ( !sslStream.IsAuthenticated ) {
                sslStream.AuthenticateAsClient( targetHost, clientCertificates, enabledProtocols, checkCertRevocation );
            }
            secure = true;
        }

        public void AuthenticateAsServer( ) {
            X509Certificate serverCert = new X509Certificate( );
            foreach ( X509Certificate cert in clientCertificates ) {
                serverCert = cert;
                break;
            }
            AuthenticateAsServer( serverCert );
        }

        public void AuthenticateAsServer( X509Certificate serverCertificate ) {
            CheckSslStream( );
            sslStream.AuthenticateAsServer( serverCertificate );
        }

        public void AuthenticateAsServer( X509Certificate serverCertificate, bool clientCertificateRequired, SslProtocols enabledProtocols, bool checkCertRevocation ) {
            CheckSslStream( );
            sslStream.AuthenticateAsServer( serverCertificate, clientCertificateRequired, enabledProtocols, checkCertRevocation );
        }

        public IAsyncResult BeginAccept( AsyncCallback callback, object @object ) {
            return AcceptInvoker.BeginInvoke( callback, @object );
        }

        public SocketEx EndAccept( IAsyncResult result ) {
            return AcceptInvoker.EndInvoke( result );
        }

        public SocketEx Accept( ) {
            var sockEx = new SocketEx { sock = sock.Accept( ) };
            return sockEx;
        }

        public SocketEx AcceptSecure( X509Certificate serverCertificate ) {
            var sockEx = new SocketEx { sock = sock.Accept( ) };
            sockEx.AuthenticateAsServer( serverCertificate );
            return sockEx;
        }

        public void Bind( EndPoint localEP ) { sock.Bind( localEP ); }

        public IAsyncResult BeginConnect( EndPoint endpoint, AsyncCallback callback, object @object ) {
            return ConnectInvoker.BeginInvoke( endpoint, callback, @object );
        }

        public void EndConnect( IAsyncResult result ) {
            ConnectInvoker.EndInvoke( result );
        }

        public void Connect( EndPoint endpoint ) {
            Connect( endpoint, false );
        }

        private void Connect( EndPoint endpoint, bool secure ) {
            if ( endpoint == null ) throw new ArgumentNullException( "endpoint" );
            sock.Connect( endpoint );
            if ( secure ) {
                this.secure = secure;
                CheckSslStream( );
            }
        }

        public IAsyncResult BeginConnectAndAuthenticate( EndPoint endpoint, string targetHost, AsyncCallback callback, object @object ) {
            return ConnectAndAuthenticateInvoker.BeginInvoke( endpoint, targetHost, callback, @object );
        }

        public void EndConnectAndAuthenticate( IAsyncResult result ) {
            ConnectAndAuthenticateInvoker.EndInvoke( result );
        }

        public void ConnectAndAuthenticate( EndPoint endpoint, string targetHost ) {
            if ( endpoint == null ) throw new ArgumentNullException( "endpoint" );
            if ( string.IsNullOrEmpty( targetHost ) ) throw new ArgumentNullException( "targetHost" );
            Connect( endpoint, true );
            AuthenticateAsClient( targetHost );
        }

        public void Close( ) {
            sock.Close( );
            if ( sslStream != null ) {
                sslStream.Close( );
                sslStream.Dispose( );
                sslStream = null;
            }
        }

        public void Dispose( ) {
            if ( !disposed ) {
                if ( sock != null ) {
                    if ( sock.Connected ) sock.Close( );
                }
                sock = null;
                if ( sslStream != null ) {
                    sslStream.Dispose( );
                    sslStream = null;
                }
                disposed = true;
                GC.SuppressFinalize( this );
            }
        }

        public void Listen( int backlog ) {
            sock.Listen( backlog );
        }

        public int Receive( byte[] buffer ) {
            if ( buffer == null ) throw new ArgumentNullException( "buffer" );
            return Receive( buffer, 0, buffer.Length, SocketFlags.None );
        }

        public int Receive( byte[] buffer, int start, int length, SocketFlags flags ) {
            if ( buffer == null ) throw new ArgumentNullException( "buffer" );
            if ( start > length ) throw new ArgumentException( "start and length conflicting" );
            if ( length < 0 ) throw new ArgumentException( "length must be greater than 0" );
            if ( secure ) {
                CheckSslStream( );
                return sslStream.Read( buffer, start, length );
            } else {
                return sock.Receive( buffer, start, length, flags );
            }
        }

        public void Send( byte[] buffer ) {
            Send( buffer, 0, buffer.Length, SocketFlags.None );
        }

        public void Send( byte[] buffer, int start, int length, SocketFlags flags ) {
            if ( buffer == null ) throw new ArgumentNullException( "buffer" );
            if ( start > length ) throw new ArgumentException( "start and length conflicting" );
            if ( length < 0 ) throw new ArgumentException( "length must be greater than 0" );
            if ( secure ) {
                CheckSslStream( );
                sslStream.Write( buffer, start, length );
            } else {
                sock.Send( buffer, start, length, flags );
            }
        }

        private void CheckSslStream( ) {
            if ( sslStream == null ) {
                switch ( clientCertMode ) {
                    case ClientCertificateMode.UseLocalMYCert: {
                        sslStream = new SslStream( new NetworkStream( sock ), true, DoServerCertificateValidation, ( sender, targetHost, localCertificates, remoteCertificate, acceptableIssuers ) => {
                            X509Store clientStore = new X509Store( StoreName.My, StoreLocation.CurrentUser );
                            foreach ( X509Certificate cert in clientStore.Certificates ) {
                                // ??? just guessing here, but if we have one, I have no clue which one would be best
                                return cert;
                            }
                            return null;
                        } );
                    }break;
                    case Net.ClientCertificateMode.UseLocalCertificateCollection: {
                        sslStream = new SslStream( new NetworkStream( sock ), true, DoServerCertificateValidation, ( sender, targetHost, localCertificates, remoteCertificate, acceptableIssuers ) => {
                            if ( clientCertificates.Count > 0 ) {
                                return clientCertificates[0];
                            }
                            return null;
                        } );
                    }break;
                    case Net.ClientCertificateMode.None:
                    default: {
                        sslStream = new SslStream( new NetworkStream( sock ), true, DoServerCertificateValidation );                        
                    }break;
                }
            }
        }

        public SocketEx( ) : this( AddressFamily.InterNetwork, SocketType.Stream, ProtocolType.Tcp ) { }

        public SocketEx( AddressFamily addressFamily, SocketType socketType, ProtocolType protocolType ) {
            sock = new Socket( addressFamily, socketType, protocolType );
            AcceptInvoker = Accept;
            ConnectInvoker = Connect;
            ConnectAndAuthenticateInvoker = ConnectAndAuthenticate;
        }
    }
}
